@@ -25,7 +25,7 @@ import math
 #     for aa_0, aa_1, score in df_blosum62.values : # sould be in [ 'aa_0', 'aa_1', 'BLOSUM62_score' ] order
 #         dict_blosum62[ aa_0, aa_1 ] = score
 
-def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', dir_folder_output = 'default', n_threads = 1, l_window_size = [ 30 ], float_thres_e_value = 1e-20, flag_use_HMM_search = False, dir_file_query_hmmdb = 'human', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_only_use_structural_properties_of_query_proteins = False, float_thres_avg_score_blosum_weighted__b_cell = 0.15, float_thres_avg_score_blosum__b_cell = 0.0, float_thres_rsa_correlation = 0.0, float_thres_avg_blosum62_score_for_mhc = 2, float_thres_min_mhc_allele_frequency = 0.5, float_thres_binding_affinities_in_nM = 500 ) :
+def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', dir_folder_output = 'default', n_threads = 1, l_window_size = [ 30 ], float_thres_e_value = 1e-20, flag_use_HMM_search = False, dir_file_query_hmmdb = 'human', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_only_use_structural_properties_of_query_proteins = False, float_thres_avg_score_blosum_weighted__b_cell = 0.15, float_thres_avg_score_blosum__b_cell = 0.0, float_thres_rsa_correlation = 0.0, float_thres_avg_blosum62_score_for_mhc = 2, float_thres_min_mhc_allele_frequency = 0.5, float_thres_binding_affinities_in_nM = 500, flag_replace_unconventional_acid_code = False, flag_use_all_gpu_devices = False ) :
     """
     The main function of Cross-Reactive-Epitope-Search-using-Structural-Properties-of-proteins (CRESSP)
     
@@ -71,7 +71,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     Package settings
     """
     name_package = 'cressp'
-    dir_remote = 'https://github.com/ahs2202/cressp2/raw/main/cressp/' # remote directory from which datafiles will be downloaded
+    dir_remote = 'https://github.com/ahs2202/cressp/raw/main/cressp/' # remote directory from which datafiles will be downloaded
     dir_folder_cressp = f"{pkg_resources.resource_filename( name_package, '' )}/" # directory of the current installed package
 
 
@@ -110,7 +110,8 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         parser.add_argument( "-b", "--float_thres_avg_blosum62_score_for_mhc", help = "(Default: 2.0) threshold for average BLOSOM62 alignment score for filtering predicted cross-reactive T-cell epitopes (cross-reactive MHC epitopes)", default = '2.0' )
         parser.add_argument( "-m", "--float_thres_min_mhc_allele_frequency", help = "(Default: 0.5) a threshold for filtering out MHC alleles with low allele frequencies. MHC alleles with allele frequency above the threshold for at least one population will be used for cross-reactive T-cell epitope prediction", default = '0.5' )
         parser.add_argument( "-a", "--float_thres_binding_affinities_in_nM", help = "(Default: 500) a threshold predicted IC50 values for filtering predicted T-cell cross-reactive epitopes. A pair of peptides whose geometric average of predicted binding affinities (IC50) values above this threshold will be removed.", default = '500' )
-
+        parser.add_argument( "-U", "--flag_replace_unconventional_acid_code", help = "(Default: False) If this flag is set, unconventional amino acids in the input protein sequences will be replaced with chemically similar amino acid. Specifically, Selenocysteine (U) to Cysteine (C), Pyrrolysine (O) to Tyrosine (Y)", action = 'store_true' )
+        parser.add_argument( "-G", "--flag_use_all_gpu_devices", help = "(Default: False) Use all available GPU devices during prediction of RSA values. When this flag is set to True, the RSA prediction might be completed faster, but all GPU memories will be occupied by Tensorflow", action = 'store_true' )
         args = parser.parse_args( )
         if args.dir_file_protein_target is None :
             print( "Required argument(s) is missing. to view help message, use -h or --help flag" )
@@ -120,6 +121,8 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         flag_use_HMM_search = args.flag_use_HMM_search
         flag_use_rcsb_pdb_only = args.flag_use_rcsb_pdb_only
         flag_only_use_structural_properties_of_query_proteins = args.flag_only_use_structural_properties_of_query_proteins
+        flag_use_all_gpu_devices = args.flag_use_all_gpu_devices
+        flag_replace_unconventional_acid_code = args.flag_replace_unconventional_acid_code
         int_number_of_proteins_in_a_batch_during_dnn_prediction = int( args.int_number_of_proteins_in_a_batch_during_dnn_prediction )
         n_threads = min( int( args.cpu ), int( multiprocessing.cpu_count( ) ) ) # max number of n_threads is the CPU number of the current local machine
         float_thres_avg_blosum62_score_for_mhc = float( args.float_thres_avg_blosum62_score_for_mhc )
@@ -181,12 +184,12 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     # handle output folder
     if dir_folder_output[ -1 ] != '/' : # last character of a directory should be '/'
         dir_folder_output += '/'
-    if os.path.exists( dir_folder_output ) :
-        print( "exiting since the given output directory already exists" )
-        if flag_usage_from_command_line_interface : sys.exit( )
-        else : return - 1
-    else :
-        os.makedirs( dir_folder_output, exist_ok = True )
+#     if os.path.exists( dir_folder_output ) :
+#         print( "exiting since the given output directory already exists" )
+#         if flag_usage_from_command_line_interface : sys.exit( )
+#         else : return - 1
+#     else :
+    os.makedirs( dir_folder_output, exist_ok = True )
     # create output folders for each task
     dir_folder_pipeline = f"{dir_folder_output}pipeline/"
     os.makedirs( dir_folder_pipeline, exist_ok = True )
@@ -217,25 +220,35 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
             name_file_without_extension, str_file_extension = name_file_without_extension.rsplit( '.', 1 )
         return name_file_without_extension
     
-    try :
-        name_file_protein_query = __Get_File_Name_Without_Extension__( dir_file_protein_query ) # retrieve file name containing the query proteins
-        dict_fasta_protein_query = FASTA_Read( dir_file_protein_query )
-        FASTA_Write( f"{dir_folder_pipeline}protein_query.fasta", dict_fasta = dict_fasta_protein_query )
-        dir_file_protein_query = f"{dir_folder_pipeline}protein_query.fasta" # set directory of fasta file to the new file location
-    except :
-        print( f"exiting due to an error while reading and moving 'dir_file_protein_query' {dir_file_protein_query}" )
-        if flag_usage_from_command_line_interface : sys.exit( )
-        else : return - 1
+    # retrieve file names
+    name_file_protein_query = __Get_File_Name_Without_Extension__( dir_file_protein_query ) # retrieve file name containing the query proteins
+    name_file_protein_target = __Get_File_Name_Without_Extension__( dir_file_protein_target ) # retrieve file name containing the query proteins
+    
+    """ check flag """
+    dir_file_flag = f"{dir_folder_pipeline}.copying_input_files_completed.flag"
+    if not os.path.exists( dir_file_flag ) :
+        
+        try :
+            dict_fasta_protein_query = FASTA_Read( dir_file_protein_query )
+            FASTA_Write( f"{dir_folder_pipeline}protein_query.fasta", dict_fasta = dict_fasta_protein_query )
+            dir_file_protein_query = f"{dir_folder_pipeline}protein_query.fasta" # set directory of fasta file to the new file location
+        except :
+            print( f"exiting due to an error while reading and moving 'dir_file_protein_query' {dir_file_protein_query}" )
+            if flag_usage_from_command_line_interface : sys.exit( )
+            else : return - 1
 
-    try :
-        name_file_protein_target = __Get_File_Name_Without_Extension__( dir_file_protein_target ) # retrieve file name containing the query proteins
-        dict_fasta_protein_target = FASTA_Read( dir_file_protein_target )
-        FASTA_Write( f"{dir_folder_pipeline}protein_target.fasta", dict_fasta = dict_fasta_protein_target )
-        dir_file_protein_target = f"{dir_folder_pipeline}protein_target.fasta" # set directory of fasta file to the new file location
-    except :
-        print( f"exiting due to an error while reading and moving 'dir_file_protein_target' {dir_file_protein_target}" )
-        if flag_usage_from_command_line_interface : sys.exit( )
-        else : return - 1
+        try :
+            dict_fasta_protein_target = FASTA_Read( dir_file_protein_target )
+            FASTA_Write( f"{dir_folder_pipeline}protein_target.fasta", dict_fasta = dict_fasta_protein_target )
+            dir_file_protein_target = f"{dir_folder_pipeline}protein_target.fasta" # set directory of fasta file to the new file location
+        except :
+            print( f"exiting due to an error while reading and moving 'dir_file_protein_target' {dir_file_protein_target}" )
+            if flag_usage_from_command_line_interface : sys.exit( )
+            else : return - 1
+            
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as newfile :
+            newfile.write( 'completed\n' )
         
     """
     Report external and internal program settings
@@ -265,7 +278,6 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         'dir_folder_pipeline_struc' : dir_folder_pipeline_struc,
         'dir_folder_pipeline_web' : dir_folder_pipeline_web,
         'dir_folder_web' : dir_folder_web,
-
         'name_file_protein_query' : name_file_protein_query,
         'name_file_protein_target' : name_file_protein_target }
     print( "cressp will be run with the following setting:" )
@@ -274,113 +286,161 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     """
     Perform BLASTP alignment
     """
-    # create blastp_db using query_protein sequences
-    dir_prefix_blastdb_protein_query = f"{dir_folder_pipeline}makeblastdb_out/protein_query"
-    os.makedirs( f"{dir_folder_pipeline}makeblastdb_out/", exist_ok = True )  
-    OS_Run( [ "makeblastdb", "-in", f"{dir_folder_pipeline}protein_query.fasta", '-dbtype', 'prot', '-parse_seqids', '-max_file_sz', '1GB', '-out', dir_prefix_blastdb_protein_query ], dir_file_stdout = f"{dir_prefix_blastdb_protein_query}.makeblastdb.stdout.txt", dir_file_stderr = f"{dir_prefix_blastdb_protein_query}.makeblastdb.stderr.txt", return_output = False ) # make blast db for protein_query
+    """ check flag """
+    dir_file_flag = f"{dir_folder_pipeline}.blastp_completed.flag"
+    if not os.path.exists( dir_file_flag ) :
 
-    # run blastp
-    dir_file_blastp_output = f'{dir_folder_pipeline}blastp.tsv'
-    OS_Run( [ 'blastp', '-query', dir_file_protein_target, '-db', dir_prefix_blastdb_protein_query, '-out', dir_file_blastp_output, '-outfmt', '6 qaccver saccver pident length mismatch gapopen qstart qend sstart send evalue bitscore btop', '-num_threads', f'{n_threads}', '-evalue', f'{float_search_thres_e_value}' ], dir_file_stdout = f"{dir_folder_pipeline}blastp.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}blastp.stderr.txt", return_output = False ) # run blastp
-    OS_Run( [ 'gzip', dir_file_blastp_output ], dir_file_stdout = f"{dir_file_blastp_output}.gzip.stdout.txt", dir_file_stderr = f"{dir_file_blastp_output}.gzip.stderr.txt", return_output = False ) # compress blastp output
-    dir_file_blastp_output += '.gz'
+        # create blastp_db using query_protein sequences
+        dir_prefix_blastdb_protein_query = f"{dir_folder_pipeline}makeblastdb_out/protein_query"
+        os.makedirs( f"{dir_folder_pipeline}makeblastdb_out/", exist_ok = True )  
+        OS_Run( [ "makeblastdb", "-in", f"{dir_folder_pipeline}protein_query.fasta", '-dbtype', 'prot', '-parse_seqids', '-max_file_sz', '1GB', '-out', dir_prefix_blastdb_protein_query ], dir_file_stdout = f"{dir_prefix_blastdb_protein_query}.makeblastdb.stdout.txt", dir_file_stderr = f"{dir_prefix_blastdb_protein_query}.makeblastdb.stderr.txt", return_output = False ) # make blast db for protein_query
 
+        # run blastp
+        dir_file_blastp_output = f'{dir_folder_pipeline}blastp.tsv'
+        OS_Run( [ 'blastp', '-query', dir_file_protein_target, '-db', dir_prefix_blastdb_protein_query, '-out', dir_file_blastp_output, '-outfmt', '6 qaccver saccver pident length mismatch gapopen qstart qend sstart send evalue bitscore btop', '-num_threads', f'{n_threads}', '-evalue', f'{float_search_thres_e_value}' ], dir_file_stdout = f"{dir_folder_pipeline}blastp.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}blastp.stderr.txt", return_output = False ) # run blastp
+        OS_Run( [ 'gzip', dir_file_blastp_output ], dir_file_stdout = f"{dir_file_blastp_output}.gzip.stdout.txt", dir_file_stderr = f"{dir_file_blastp_output}.gzip.stderr.txt", return_output = False ) # compress blastp output
+        dir_file_blastp_output += '.gz'
+        
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as newfile :
+            newfile.write( 'completed\n' )
 
     """
     Perform HMMER alignment using a given HMM profile DB (using a couple of HMM profile DBs would be also helpful in the near future)
     """
     # run hmmsearch according to 'flag_use_HMM_search' flag
     if flag_use_HMM_search : 
-        dir_file_hmmsearch_output = f'{dir_folder_pipeline}hmmsearch.out'
-        OS_Run( [ 'hmmsearch', '-o', dir_file_hmmsearch_output, '--acc', '--notextw', '--cpu', f'{n_threads}', '-E', f'{float_search_thres_e_value}', dir_file_query_hmmdb, dir_file_protein_target ], dir_file_stdout = f"{dir_folder_pipeline}hmmsearch.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}hmmsearch.stderr.txt", return_output = False ) # run hmmsearch
+        """ check flag """
+        dir_file_flag = f"{dir_folder_pipeline}.hmmsearch_completed.flag"
+        if not os.path.exists( dir_file_flag ) :
+            dir_file_hmmsearch_output = f'{dir_folder_pipeline}hmmsearch.out'
+            OS_Run( [ 'hmmsearch', '-o', dir_file_hmmsearch_output, '--acc', '--notextw', '--cpu', f'{n_threads}', '-E', f'{float_search_thres_e_value}', dir_file_query_hmmdb, dir_file_protein_target ], dir_file_stdout = f"{dir_folder_pipeline}hmmsearch.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}hmmsearch.stderr.txt", return_output = False ) # run hmmsearch
+            """ set flag """
+            with open( dir_file_flag, 'w' ) as newfile :
+                newfile.write( 'completed\n' )
 
     """
     Combine BLASTP and HMMSEARCH outputs 
     """
     dir_file_matched = f'{dir_folder_pipeline}matched.tsv.gz'
-    dir_file_matched_write_complete_flag = f"{dir_file_matched}.write_completed.flag" # a flag indicating the write operation of the file has been completed.
+    
+    """ check flag """
+    dir_file_flag = f"{dir_file_matched}.write_completed.flag"
+    if not os.path.exists( dir_file_flag ) :
+        # load blastp result
+        dict_qacc_to_seq = FASTA_Read( dir_file_protein_target ) # read query protein sequences
+        dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_qacc_to_seq[ header ] ) for header in list( dict_qacc_to_seq ) )
+        df_blastp = BLAST_Read( dir_file_blastp_output, dict_qaccver_to_seq = dict_qacc_to_seq ) 
+        df_blastp = df_blastp[ [ 'saccver', 'qaccver', 'sstart', 'send', 'qstart', 'qend', 'subject_seq_aligned', 'query_seq_aligned', 'evalue', 'pident' ] ]
+        df_blastp.pident = df_blastp.pident / 100
+        df_blastp.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ]
+        df_blastp[ 'source' ] = 'blastp'
 
-    # load blastp result
-    dict_qacc_to_seq = FASTA_Read( dir_file_protein_target ) # read query protein sequences
-    dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_qacc_to_seq[ header ] ) for header in list( dict_qacc_to_seq ) )
-    df_blastp = BLAST_Read( dir_file_blastp_output, dict_qaccver_to_seq = dict_qacc_to_seq ) 
-    df_blastp = df_blastp[ [ 'saccver', 'qaccver', 'sstart', 'send', 'qstart', 'qend', 'subject_seq_aligned', 'query_seq_aligned', 'evalue', 'pident' ] ]
-    df_blastp.pident = df_blastp.pident / 100
-    df_blastp.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ]
-    df_blastp[ 'source' ] = 'blastp'
+        # load hmmer result according to 'flag_use_HMM_search' flag
+        if flag_use_HMM_search : 
+            df = HMMER_HMMSEARCH_Read_output( dir_file_hmmsearch_output )
+            dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_fasta_protein_query[ header ] ) for header in list( dict_fasta_protein_query ) )
+            l_query_alignment = list( ) # replace query consensus sequence with query sequence
+            for query_accession, query_alignment, query_start, query_end in df[ [ 'query_accession', 'query_alignment', 'query_start', 'query_end' ] ].values :
+                query_seq = dict_qacc_to_seq[ query_accession ][ query_start - 1 : query_end ] # retrive a subsequence of query sequence
+                l_subsequence = list( )
+                int_start = 0
+                for subsequence in query_alignment.split( '.' ) :
+                    int_subsequence_length = len( subsequence )
+                    l_subsequence.append( query_seq[ int_start : int_start + int_subsequence_length ] )
+                    int_start += int_subsequence_length
+                l_query_alignment.append( '-'.join( l_subsequence ) )
+            df[ 'query_alignment' ] = l_query_alignment
+            df[ 'target_alignment' ] = df.target_alignment.str.upper( ) # target alignment string contains amino acids in lower cases when alignment confidence is low, and it is convenient to convert them to upper characters
+            df_hmmer = df
+            df_hmmer = df_hmmer[ [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'conditional_Evalue', 'accuracy' ] ] # subset common columns
+            df_hmmer.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ] # rename columns
+            df_hmmer[ 'source' ] = 'hmmer' 
 
-    # load hmmer result according to 'flag_use_HMM_search' flag
-    if flag_use_HMM_search : 
-        df = HMMER_HMMSEARCH_Read_output( dir_file_hmmsearch_output )
-        dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_fasta_protein_query[ header ] ) for header in list( dict_fasta_protein_query ) )
-        l_query_alignment = list( ) # replace query consensus sequence with query sequence
-        for query_accession, query_alignment, query_start, query_end in df[ [ 'query_accession', 'query_alignment', 'query_start', 'query_end' ] ].values :
-            query_seq = dict_qacc_to_seq[ query_accession ][ query_start - 1 : query_end ] # retrive a subsequence of query sequence
-            l_subsequence = list( )
-            int_start = 0
-            for subsequence in query_alignment.split( '.' ) :
-                int_subsequence_length = len( subsequence )
-                l_subsequence.append( query_seq[ int_start : int_start + int_subsequence_length ] )
-                int_start += int_subsequence_length
-            l_query_alignment.append( '-'.join( l_subsequence ) )
-        df[ 'query_alignment' ] = l_query_alignment
-        df[ 'target_alignment' ] = df.target_alignment.str.upper( ) # target alignment string contains amino acids in lower cases when alignment confidence is low, and it is convenient to convert them to upper characters
-        df_hmmer = df
-        df_hmmer = df_hmmer[ [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'conditional_Evalue', 'accuracy' ] ] # subset common columns
-        df_hmmer.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ] # rename columns
-        df_hmmer[ 'source' ] = 'hmmer' 
+        df_matched = pd.concat( [ df_hmmer, df_blastp ], ignore_index = True ) if flag_use_HMM_search else df_blastp
+        df_matched.to_csv( dir_file_matched, sep = '\t', index = False )
 
-    df_matched = pd.concat( [ df_hmmer, df_blastp ], ignore_index = True ) if flag_use_HMM_search else df_blastp
-    df_matched.to_csv( dir_file_matched, sep = '\t', index = False )
-
-    # write a flag (file) indicating the writing operation was completed.
-    with open( dir_file_matched_write_complete_flag, 'w' ) as file :
-        file.write( f"search results were written at {TIME_GET_timestamp( )}" )
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as file :
+            file.write( f"search results were written at {TIME_GET_timestamp( )}" )
 
     """
     Estimate structural properties of proteins 
     """
-    # use previously calculated structural properties when the default query proteins were used
-    if flag_default_protein_query_was_used :
-        shutil.copyfile( f'{dir_folder_cressp}data/human/uniprot.tsv.gz', f'{dir_folder_pipeline}protein_query.tsv.gz' ) 
-        # use previously calculated alignment between human proteins and RCSB_PDB structures
-        PKG.Download_Data( "data/human/uniprot.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz", dir_remote, name_package ) # download data (alignment between human proteins and RCSB_PDB structures)
-        shutil.copyfile( f'{dir_folder_cressp}data/human/uniprot.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz', f'{dir_folder_pipeline}struc/protein_query.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz' ) 
     
-    # estimate structural properties
-    for name_file in [ 'protein_target' ] if flag_default_protein_query_was_used else [ 'protein_target', 'protein_query' ] : # skip prediction of query proteins if default query proteins are used
-        """ check flag """
-        dir_file_flag = f"{dir_folder_pipeline}{name_file}.tsv.gz.completed.flag"
-        if not os.path.exists( dir_file_flag ) :
-            Estimate_structural_property( f'{dir_folder_pipeline}{name_file}.fasta', n_threads, dir_folder_pipeline, dir_folder_pipeline_temp, flag_use_rcsb_pdb_only, int_number_of_proteins_in_a_batch_during_dnn_prediction )
+    """ check flag """
+    dir_file_flag = f"{dir_file_matched}.rsa_estimation_complated.flag"
+    if not os.path.exists( dir_file_flag ) :
+        # use previously calculated structural properties when the default query proteins were used
+        if flag_default_protein_query_was_used :
+            shutil.copyfile( f'{dir_folder_cressp}data/human/uniprot.tsv.gz', f'{dir_folder_pipeline}protein_query.tsv.gz' ) 
+            # use previously calculated alignment between human proteins and RCSB_PDB structures
+            PKG.Download_Data( "data/human/uniprot.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz", dir_remote, name_package ) # download data (alignment between human proteins and RCSB_PDB structures)
+            shutil.copyfile( f'{dir_folder_cressp}data/human/uniprot.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz', f'{dir_folder_pipeline}struc/protein_query.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz' ) 
 
-            """ set flag """
-            with open( dir_file_flag, 'w' ) as newfile :
-                newfile.write( 'completed\n' )
+        # estimate structural properties
+        for name_file in [ 'protein_target' ] if flag_default_protein_query_was_used else [ 'protein_target', 'protein_query' ] : # skip prediction of query proteins if default query proteins are used
+            """ check flag_2 """
+            dir_file_flag_2 = f"{dir_folder_pipeline}{name_file}.tsv.gz.completed.flag"
+            if not os.path.exists( dir_file_flag_2 ) :
+                Estimate_structural_property( f'{dir_folder_pipeline}{name_file}.fasta', n_threads, dir_folder_pipeline, dir_folder_pipeline_temp, flag_use_rcsb_pdb_only, int_number_of_proteins_in_a_batch_during_dnn_prediction, flag_use_all_gpu_devices )
+
+                """ set flag_2 """
+                with open( dir_file_flag_2, 'w' ) as newfile :
+                    newfile.write( 'completed\n' )
+                    
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as newfile :
+            newfile.write( 'completed\n' )
 
 
     """
     Calculate B-cell epitope similarity scores based on structural properties of proteins 
     """
-    Predict_B_cell_cross_reactivity( dir_folder_pipeline, dir_folder_pipeline_temp, n_threads, l_window_size, float_thres_e_value, flag_only_use_structural_properties_of_query_proteins, float_thres_avg_score_blosum_weighted__b_cell, float_thres_avg_score_blosum__b_cell, float_thres_rsa_correlation )
+    
+    """ check flag """
+    dir_file_flag = f"{dir_file_matched}.predicting_B_cell_cross_reactivity_completed.flag"
+    if not os.path.exists( dir_file_flag ) :
+        
+        Predict_B_cell_cross_reactivity( dir_folder_pipeline, dir_folder_pipeline_temp, n_threads, l_window_size, float_thres_e_value, flag_only_use_structural_properties_of_query_proteins, float_thres_avg_score_blosum_weighted__b_cell, float_thres_avg_score_blosum__b_cell, float_thres_rsa_correlation )
+        
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as newfile :
+            newfile.write( 'completed\n' )
     
     """ 
     Calculate T-cell epitope similarity scores based on BLOSUM62 scores and predicted binding affinity scores.
     """
     
-    Predict_T_cell_cross_reactivity( dir_folder_pipeline, float_thres_avg_blosum62_score_for_mhc, float_thres_min_mhc_allele_frequency, float_thres_binding_affinities_in_nM )
-
+    """ check flag """
+    dir_file_flag = f"{dir_file_matched}.predicting_T_cell_cross_reactivity_completed.flag"
+    if not os.path.exists( dir_file_flag ) :
+        
+        Predict_T_cell_cross_reactivity( dir_folder_pipeline, float_thres_avg_blosum62_score_for_mhc, float_thres_min_mhc_allele_frequency, float_thres_binding_affinities_in_nM, flag_replace_unconventional_acid_code )
+        
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as newfile :
+            newfile.write( 'completed\n' )
     
     """
     Further process and export data for visualization using a web application
     """
-    # combine results of all 'window_size' values
-    df = pd.concat( list( pd.read_csv( dir_file, sep = '\t', low_memory = False ) for dir_file in glob.glob( f"{dir_folder_pipeline}b_cell.subsequence__window_size_*.tsv.gz" ) ), ignore_index = True )
-    df.to_csv( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", sep = '\t', index = False )
-    del df
-    # prepare data for web application using the combined subsequence
-    # copy data for web application and encode using base64 encoding, and write metadata
-    Prepare_data_for_web_application( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", f"{dir_folder_pipeline}t_cell.mhc_binding.tsv.gz", dict_cressp_setting )
+    
+    """ check flag """
+    dir_file_flag = f"{dir_file_matched}.prepare_data_for_web_application_completed.flag"
+    if not os.path.exists( dir_file_flag ) :
+
+        # combine results of all 'window_size' values
+        df = pd.concat( list( pd.read_csv( dir_file, sep = '\t', low_memory = False ) for dir_file in glob.glob( f"{dir_folder_pipeline}b_cell.subsequence__window_size_*.tsv.gz" ) ), ignore_index = True )
+        df.to_csv( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", sep = '\t', index = False )
+        del df
+        # prepare data for web application using the combined subsequence
+        # copy data for web application and encode using base64 encoding, and write metadata
+        Prepare_data_for_web_application( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", f"{dir_folder_pipeline}t_cell.mhc_binding.tsv.gz", dict_cressp_setting )
+        
+        """ set flag """
+        with open( dir_file_flag, 'w' ) as newfile :
+            newfile.write( 'completed\n' )
     
 
 
@@ -441,4 +501,4 @@ def Encode_Structural_Properties( dict_sp, name_col_for_identifying_protein = 'i
     
 if __name__ == "__main__" :
     cressp( ) # run CRESSP at the top-level environment
-    
\ No newline at end of file
+    
