@@ -150,8 +150,269 @@ def Combine_Result__Transfer_DSSP_Structural_Property_Through_BLAST( str_uuid, d
     OS_FILE_Combine_Files_in_order( glob.glob( f"{dir_temp}{str_uuid}*.tsv.gz.transferred.tsv.gz" ), f"{dir_folder_pipeline_struc}{name_file}_transferred_to_{name_dataset}.tsv.gz", flag_use_header_from_first_file = True, remove_n_lines = 1, overwrite_existing_file = True )
 
 
+def __Combine_Structural_Properties_and_Predict_Remaining_Residues__( dir_file_protein, dir_folder_pipeline = None, dir_folder_pipeline_temp = '/tmp/', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_use_all_gpu_devices = False ) :
+    
+    """
+    Parse arguments
+    """
+    # directories of file and folders
+    dir_file_protein = os.path.abspath( dir_file_protein )
+    dir_folder_pipeline = os.path.abspath( dir_folder_pipeline )
+    dir_folder_pipeline_temp = os.path.abspath( dir_folder_pipeline_temp )
+    if dir_folder_pipeline[ -1 ] != '/' : # last character of a directory should be '/'
+        dir_folder_pipeline += '/'
+    if dir_folder_pipeline_temp[ -1 ] != '/' : # last character of a directory should be '/'
+        dir_folder_pipeline_temp += '/'
+    
+    # fixed arguments
+    float_search_thres_e_value = 30 # set e-value threshold for search (fixed for maximum sensitivity)
 
-def Estimate_structural_property( dir_file_protein, n_threads, dir_folder_pipeline = None, dir_folder_pipeline_temp = '/tmp/', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000 ) :
+    # By default, set 'dir_folder_pipeline' as the folder where the given protein fasta file is located
+    if dir_folder_pipeline is None :
+        dir_folder_pipeline = dir_file_protein.rsplit( '/', 1 )[ 0 ] + '/'
+    name_file = dir_file_protein.rsplit( '/', 1 )[ 1 ].rsplit( '.', 1 )[ 0 ] # retrieve name of file excluding file extension
+    dir_file_protein_property = f"{dir_folder_pipeline}{name_file}.tsv.gz" # define output file (a file containing protein sequences and their estimated structural properties)
+    dir_folder_pipeline_struc = f'{dir_folder_pipeline}struc/' # create a working directory of estimating structural properties
+    os.makedirs( dir_folder_pipeline_struc, exist_ok = True )
+    
+    """
+    Package settings
+    """
+    name_package = 'cressp'
+    dir_remote = 'https://github.com/ahs2202/cressp/raw/main/cressp/' # remote directory from which datafiles will be downloaded
+    dir_folder_cressp = f"{pkg_resources.resource_filename( name_package, '' )}/" # directory of the current installed package
+    
+    """ read input proteins """
+    dict_fasta_protein = FASTA_Read( dir_file_protein )
+    dict_header = dict( ( h.split( ' ', 1 )[ 0 ], h ) for h in dict_fasta_protein ) # retrieve a dictionary of header (key = id_protein, value = header)
+    dict_fasta_protein = dict( ( h.split( ' ', 1 )[ 0 ], dict_fasta_protein[ h ] ) for h in dict_fasta_protein ) # retrieve sequence_id by spliting the header at the first space (to makes sequence_id consistent with that used with blastp)
+    
+    """ 
+    Load and Combine Structural Properties
+    """
+    # 2021-05-29 17:25:14 
+
+    def __Parse_Structural_Properties__( df_sp, int_datatype ) :
+        """
+        # 2021-05-03 16:18:16 
+        parse structural properties with typical parameters for parsing ascii-encoded structural properties and typical column names  
+        Also, initializes a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
+        'int_datatype' : integer identifying the current dataset. Used for initialization of a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
+        """
+        dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
+        dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
+        dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
+        dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )
+
+        dict_sp = dict( )
+        dict_sp[ 'acc' ] = ASCII_Decode( df_sp.set_index( 'header' ).rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1.to_dict( ), ** dict_kw_rsa ) # decode mkdssp outputs of RCSB PDB data
+        dict_sp[ 'phi' ] = ASCII_Decode( df_sp.set_index( 'header' )[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
+        dict_sp[ 'psi' ] = ASCII_Decode( df_sp.set_index( 'header' )[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
+        dict_sp[ 'ss8' ] = ASCII_Decode( df_sp.set_index( 'header' )[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].to_dict( ), ** dict_kw_ss8 )
+        dict_fasta = df_sp.set_index( 'header' )[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
+        dict_sp[ 'structure_id' ] = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
+
+        # initialization of a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
+        dict_datatype_acc = dict( )
+        for h in dict_sp[ 'acc' ] :
+            arr = np.zeros_like( dict_sp[ 'acc' ][ h ] ) # default = 0 (predicted structural property)
+            arr[ ~ np.isnan( dict_sp[ 'acc' ][ h ] ) ] = int_datatype
+            dict_datatype_acc[ h ] = arr
+        dict_sp[ 'datatype_acc' ] = dict_datatype_acc
+        return dict_sp
+
+    def __Combine_Structural_Properties__( dict_sp_1, dict_sp_2 ) :
+        """
+        # 2021-05-29 16:12:09 
+        Combine structural properties from the two different sources. 
+        Structural properties from the source with the lower priority will be overwritten with that of the higher priority
+
+        'df_sp_1' : structural properties of the lower priority
+        'df_sp_2' : structural properties with the higher priority
+        """
+        dict_sp_combined = dict( )
+
+        for str_datatype in [ 'acc', 'phi', 'psi', 'ss8', 'structure_id', 'datatype_acc' ] :
+            dict_arr_1 = dict_sp_1[ str_datatype ]
+            dict_arr_2 = dict_sp_2[ str_datatype ]
+            dict_arr_combined = dict( )
+            for h in set( dict_arr_1 ).union( set( dict_arr_2 ) ) :
+                # retrieve length of the protein sequence
+                len_seq = 0
+                for dict_arr in [ dict_arr_1, dict_arr_2 ] : 
+                    if h in dict_arr :
+                        len_seq = len( dict_arr[ h ] )
+
+                arr_combined = np.full( len_seq, np.nan, dtype = object ) if str_datatype == 'structure_id' else np.full( len_seq, np.nan ) # initialize array containing combined data
+                # update data in the order of increasing priorities (so that data with the highest priority is updated last and thus not overwritten by any other data)
+                for dict_arr in [ dict_arr_2 ] if str_datatype == 'structure_id' else [ dict_arr_1, dict_arr_2 ] : # for 'structure_id', only use structure_id of the database of the higher priority ('RCSB_PDB' to be exact) 
+                    if h in dict_arr :
+                        arr_current = dict_arr[ h ]
+                        mask_valid = ~ pd.isnull( arr_current ) if str_datatype == 'structure_id' else ~ np.isnan( arr_current )
+                        arr_combined[ mask_valid ] = arr_current[ mask_valid ] # update data
+                dict_arr_combined[ h ] = arr_combined
+            dict_sp_combined[ str_datatype ] = dict_arr_combined
+        return dict_sp_combined
+
+    df_sp = pd.read_csv( f"{dir_folder_pipeline_struc}{name_file}_transferred_to_rcsb_pdb.tsv.gz", sep = '\t' )
+    dict_sp = __Parse_Structural_Properties__( df_sp, 2 ) # parse structural data from rcsb_pdb
+
+    if not flag_use_rcsb_pdb_only :
+        df_sp_swiss = pd.read_csv( f"{dir_folder_pipeline_struc}{name_file}_transferred_to_swiss_model.tsv.gz", sep = '\t' )
+        dict_sp_swiss = __Parse_Structural_Properties__( df_sp_swiss, 1 ) # parse structural data from swiss-model
+
+        dict_sp = __Combine_Structural_Properties__( dict_sp_swiss, dict_sp ) # combine structural property data from RCSB_PDB and SWISS-MODEL
+
+    ''' initialize arrays of proteins with no aligned structures '''
+    for id_protein in dict_fasta_protein :
+        len_seq = len( dict_fasta_protein[ id_protein ] ) # retrieve length of protein
+        for str_datatype in [ 'acc', 'phi', 'psi', 'ss8', 'structure_id', 'datatype_acc' ] :
+            if id_protein not in dict_sp[ str_datatype ] :
+                ''' set default array for each datatype '''
+                if str_datatype == 'datatype_acc' :
+                    arr = np.full( len_seq, 0, dtype = float )
+                elif str_datatype == 'structure_id' :
+                    arr = np.full( len_seq, np.nan, dtype = object )
+                else :
+                    arr = np.full( len_seq, np.nan )
+                dict_sp[ str_datatype ][ id_protein ] = arr
+
+    """
+    Prediction of Structural Properties Using Machine-Learning Models
+    """
+    # 2021-05-29 23:41:46 
+    # use machine-learning model to predict structural properties
+    ''' settings '''
+    int_length_flanking = 74 # number of protein residues flanking the residue whose structural property values are predicted
+
+    ''' download models '''
+    file_name_model_rsa = "RSA.Conv1D.5_large_and_wide_dense.epoch_082.val_loss_0.1185.hdf5" # name of machine learning model used for structural property prediction (RSA)
+    file_name_model_ss8 = "SS8.Conv1D.5_large_and_wide_dense.epoch_031.val_loss_0.7378.hdf5" # name of machine learning model used for structural property prediction (SS8)
+
+    PKG.Download_Data( f"structural_property_estimation/{file_name_model_rsa}", dir_remote, name_package ) # download data
+    PKG.Download_Data( f"structural_property_estimation/{file_name_model_ss8}", dir_remote, name_package ) # download data
+
+    ''' prepare encoding '''
+    def Index_List( l, start = 0 ) :
+        ''' # 2021-05-06 22:32:53 
+        index a given list containing hashable element with integer index starting with 'start' (default: 0)
+        can be used for 'one-hot encoding'
+        '''
+        return dict( ( e, i ) for e, i in zip( l, np.arange( start, len( l ) + start, 1, dtype = int ) ) )
+
+    # input structure : amino-acids (one-hot encoding) + ss8 (one-hot encoding) + rsa (0~1)
+    int_index_start_amino_acid = 0
+    l_amino_acids = [ 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y' ]
+    dict_amino_acid_to_index = Index_List( l_amino_acids, int_index_start_amino_acid ) # one-hot encoding dictionary for 20 amino acids
+    int_index_start_ss8 = len( l_amino_acids )
+    l_ss8 = [ 'G', 'H', 'I', 'E', 'B', 'T', 'S', 'C' ]
+    dict_ss8_to_index = Index_List( l_ss8, int_index_start_ss8 ) # one-hot encoding dictionary for 20 amino acids
+    index_rsa = len( l_amino_acids ) + len( l_ss8 ) # index of rsa data in the input array
+
+
+    ''' predict and update structural properties using ML models for each batch '''
+    import tensorflow as tf
+    from tensorflow.keras import layers
+
+    l_id_protein_all = list( dict_sp[ 'acc' ] ) # retrieve a list containing all 'id_protein'
+    n_batches = int( np.ceil( len( l_id_protein_all ) / int_number_of_proteins_in_a_batch_during_dnn_prediction ) ) # number of batches
+    ''' for each batch '''
+    for l_id_protein in LIST_Split( l_id_protein_all, n_split = n_batches ) :
+        ''' perform encoding for the current batch '''
+        # 'l_id_protein': id_protein -> integer encoding to reduce memory usage
+        # perform encoding and record metadata of each position
+        x_data = [ ]
+        meta_data = [ ]
+
+        for index_protein, id_protein in enumerate( l_id_protein ) :
+            seq = dict_fasta_protein[ id_protein ]
+            length = len( seq ) # retrieve the length of the protein
+
+            arr_acc = dict_sp[ 'acc' ][ id_protein ]
+            arr_ss8 = dict_sp[ 'ss8' ][ id_protein ]
+
+            ''' one-hot encoding (+ RSA) of the current protein '''
+            arr_protein = np.zeros( ( length, len( l_amino_acids ) + len( l_ss8 ) + 1 ), dtype = np.float32 ) # initialize one-hot encoding of the proteinarray
+            for i in range( length ) :
+                # one-hot encoding of amino acid
+                aa = seq[ i ] # retrieve amino acid
+                if aa in dict_amino_acid_to_index : # invalid amino acid will be considered as a missing value (empty row)
+                    arr_protein[ i, dict_amino_acid_to_index[ aa ] ] = 1 
+                # one-hot encoding of ss8
+                if not np.isnan( arr_ss8[ i ] ) : # missing value = empty row
+                    arr_protein[ i, dict_ss8_to_index[ l_ss8[ int( arr_ss8[ i ] ) ] ] ] = 1
+                # encode rsa information
+                if not np.isnan( arr_acc[ i ] ) : # missing value = empty row
+                    arr_protein[ i, index_rsa ] = arr_acc[ i ]
+
+            ''' prepare one-hot encoding input array for each missing position '''
+            ''' assumes positions of missing values in 'arr_acc' are identical to those of 'arr_ss8' '''
+            for int_pos in np.where( np.isnan( arr_acc ) )[ 0 ] : 
+                start = max( 0, int_pos - int_length_flanking )
+                end = min( length, int_pos + int_length_flanking + 1 )
+
+                arr = np.zeros( ( int_length_flanking + 1 + int_length_flanking, len( l_amino_acids ) + len( l_ss8 ) + 1 ), dtype = np.float32 ) # initialize input array
+                start_input = max( 0, int_length_flanking - int_pos ) # index of start position of the input array
+                arr[ start_input : start_input + end - start ] = arr_protein[ start : end ]
+                x_data.append( arr )
+                meta_data.append( [ index_protein, int_pos ] )
+        x_data = np.array( x_data, dtype = np.float32 )
+
+        ''' predict and update structural properties using ML models for the current batch '''
+        for dir_file_model, str_datatype in zip( [ f"{dir_folder_cressp}structural_property_estimation/{file_name_model_rsa}", f"{dir_folder_cressp}structural_property_estimation/{file_name_model_ss8}" ], [ 'acc', 'ss8' ] ) :
+
+            model = tf.keras.models.load_model( dir_file_model )
+
+            try :
+                with tf.distribute.MirroredStrategy( ).scope( ) if flag_use_all_gpu_devices else open( f"{dir_folder_cressp}structural_property_estimation/{UUID( )}.flag", 'w' ) : # distributed calculation if 'flag_use_all_gpus' is True
+                    y_pred = model.predict( x_data ) # predict structural properties using a ML model
+                    y_pred = y_pred.argmax( axis = 1 ) if str_datatype == 'ss8' else y_pred # multiclass classification for 'ss8' and regression for 'acc'
+            except tf.errors.InternalError : # when out of memory error occurs during training, use CPU instead
+                os.environ[ 'CUDA_VISIBLE_DEVICES' ] = '-1' # forcing tensorflow to use CPU
+                y_pred = model.predict( x_data ) # predict structural properties using a ML model
+                y_pred = y_pred.argmax( axis = 1 ) if str_datatype == 'ss8' else y_pred # multiclass classification for 'ss8' and regression for 'acc'
+
+            for i in range( len( y_pred ) ) :
+                index_protein, int_pos = meta_data[ i ] # parse metadata
+                id_protein = l_id_protein[ index_protein ]
+                dict_sp[ str_datatype ][ id_protein ][ int_pos ] = y_pred[ i ] # update structural property data with predicted data
+
+    """
+    Save Result
+    """
+    def __Encode_Structural_Properties__( dict_sp, dict_fasta_protein ) :
+        """
+        # 2021-05-29 23:56:59 
+        Compose a dataframe containing encoded structural properties using given 'dict_sp' and 'dict_fasta_protein'
+
+        'dict_sp' : dictionary containing structural properties
+        'dict_fasta_protein' : dictionary containing protein sequences
+
+        """
+        dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
+        dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
+        dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
+        dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )
+
+        ''' initialize dataframe with protein sequences '''
+        df_sp = pd.Series( dict_fasta_protein, name = 'seq' ).reset_index( ).rename( columns = { 'index' : 'id_protein' } ).set_index( 'id_protein' )
+
+        # encode combined structural properties into ASCII strings using ASCII encoding
+        df_sp[ 'rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1' ] = pd.Series( ASCII_Encode( dict_sp[ 'acc' ], ** dict_kw_rsa ) )
+        df_sp[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ] = pd.Series( ASCII_Encode( dict_sp[ 'phi' ], ** dict_kw_torsion_angle ) )
+        df_sp[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ] = pd.Series( ASCII_Encode( dict_sp[ 'psi' ], ** dict_kw_torsion_angle ) )
+        df_sp[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ] = pd.Series( ASCII_Encode( dict_sp[ 'ss8' ], ** dict_kw_ss8 ) )
+        df_sp[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ] = pd.Series( ASCII_Encode( dict_sp[ 'datatype_acc' ], ** dict_kw_datatype ) )
+        df_sp[ 'structure_id___redundancy_reduced' ] = pd.Series( dict( ( h, Encode_List_of_Strings( dict_sp[ 'structure_id' ][ h ] ) ) for h in dict_sp[ 'structure_id' ] ) )
+        df_sp.reset_index( drop = False, inplace = True )
+        return df_sp
+
+    df_sp = __Encode_Structural_Properties__( dict_sp, dict_fasta_protein ) # encode structural properties into a dataframe
+    df_sp[ 'fasta_header' ] = list( dict_header[ i ] for i in df_sp.id_protein.values ) # retrieve headers
+    df_sp.to_csv( f"{dir_folder_pipeline}{name_file}.tsv.gz", sep = '\t', index = False ) # save structural properties of input proteins as a tabular file
+        
+
+def Estimate_structural_property( dir_file_protein, n_threads, dir_folder_pipeline = None, dir_folder_pipeline_temp = '/tmp/', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_use_all_gpu_devices = False ) :
     """
     # 2021-05-31 15:13:13 
     Estimate structural properties of given proteins, and write a tsv file containing structural properties of the proteins.
@@ -187,7 +448,7 @@ def Estimate_structural_property( dir_file_protein, n_threads, dir_folder_pipeli
     Package settings
     """
     name_package = 'cressp'
-    dir_remote = 'https://github.com/ahs2202/cressp2/raw/main/cressp/' # remote directory from which datafiles will be downloaded
+    dir_remote = 'https://github.com/ahs2202/cressp/raw/main/cressp/' # remote directory from which datafiles will be downloaded
     dir_folder_cressp = f"{pkg_resources.resource_filename( name_package, '' )}/" # directory of the current installed package
 
 
@@ -300,229 +561,14 @@ def Estimate_structural_property( dir_file_protein, n_threads, dir_folder_pipeli
             with open( dir_file_flag, 'w' ) as newfile :
                 newfile.write( 'completed\n' )
 
-                
+    """ 
+    Combine Structural Properties and Predict Structural Properties of Remaining Residues
+    """  
     """ check flag """
     dir_file_flag = f"{dir_folder_pipeline}{name_file}.tsv.gz.completed.flag"
     if not os.path.exists( dir_file_flag ) :
-        
-        """ 
-        Load and Combine Structural Properties
-        """
-        # 2021-05-29 17:25:14 
-
-        def __Parse_Structural_Properties__( df_sp, int_datatype ) :
-            """
-            # 2021-05-03 16:18:16 
-            parse structural properties with typical parameters for parsing ascii-encoded structural properties and typical column names  
-            Also, initializes a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
-            'int_datatype' : integer identifying the current dataset. Used for initialization of a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
-            """
-            dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
-            dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
-            dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
-            dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )
-
-            dict_sp = dict( )
-            dict_sp[ 'acc' ] = ASCII_Decode( df_sp.set_index( 'header' ).rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1.to_dict( ), ** dict_kw_rsa ) # decode mkdssp outputs of RCSB PDB data
-            dict_sp[ 'phi' ] = ASCII_Decode( df_sp.set_index( 'header' )[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
-            dict_sp[ 'psi' ] = ASCII_Decode( df_sp.set_index( 'header' )[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
-            dict_sp[ 'ss8' ] = ASCII_Decode( df_sp.set_index( 'header' )[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].to_dict( ), ** dict_kw_ss8 )
-            dict_fasta = df_sp.set_index( 'header' )[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
-            dict_sp[ 'structure_id' ] = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
-
-            # initialization of a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
-            dict_datatype_acc = dict( )
-            for h in dict_sp[ 'acc' ] :
-                arr = np.zeros_like( dict_sp[ 'acc' ][ h ] ) # default = 0 (predicted structural property)
-                arr[ ~ np.isnan( dict_sp[ 'acc' ][ h ] ) ] = int_datatype
-                dict_datatype_acc[ h ] = arr
-            dict_sp[ 'datatype_acc' ] = dict_datatype_acc
-            return dict_sp
-
-        def __Combine_Structural_Properties__( dict_sp_1, dict_sp_2 ) :
-            """
-            # 2021-05-29 16:12:09 
-            Combine structural properties from the two different sources. 
-            Structural properties from the source with the lower priority will be overwritten with that of the higher priority
-
-            'df_sp_1' : structural properties of the lower priority
-            'df_sp_2' : structural properties with the higher priority
-            """
-            dict_sp_combined = dict( )
-
-            for str_datatype in [ 'acc', 'phi', 'psi', 'ss8', 'structure_id', 'datatype_acc' ] :
-                dict_arr_1 = dict_sp_1[ str_datatype ]
-                dict_arr_2 = dict_sp_2[ str_datatype ]
-                dict_arr_combined = dict( )
-                for h in set( dict_arr_1 ).union( set( dict_arr_2 ) ) :
-                    # retrieve length of the protein sequence
-                    len_seq = 0
-                    for dict_arr in [ dict_arr_1, dict_arr_2 ] : 
-                        if h in dict_arr :
-                            len_seq = len( dict_arr[ h ] )
-
-                    arr_combined = np.full( len_seq, np.nan, dtype = object ) if str_datatype == 'structure_id' else np.full( len_seq, np.nan ) # initialize array containing combined data
-                    # update data in the order of increasing priorities (so that data with the highest priority is updated last and thus not overwritten by any other data)
-                    for dict_arr in [ dict_arr_2 ] if str_datatype == 'structure_id' else [ dict_arr_1, dict_arr_2 ] : # for 'structure_id', only use structure_id of the database of the higher priority ('RCSB_PDB' to be exact) 
-                        if h in dict_arr :
-                            arr_current = dict_arr[ h ]
-                            mask_valid = ~ pd.isnull( arr_current ) if str_datatype == 'structure_id' else ~ np.isnan( arr_current )
-                            arr_combined[ mask_valid ] = arr_current[ mask_valid ] # update data
-                    dict_arr_combined[ h ] = arr_combined
-                dict_sp_combined[ str_datatype ] = dict_arr_combined
-            return dict_sp_combined
-
-        df_sp = pd.read_csv( f"{dir_folder_pipeline_struc}{name_file}_transferred_to_rcsb_pdb.tsv.gz", sep = '\t' )
-        dict_sp = __Parse_Structural_Properties__( df_sp, 2 ) # parse structural data from rcsb_pdb
-
-        if not flag_use_rcsb_pdb_only :
-            df_sp_swiss = pd.read_csv( f"{dir_folder_pipeline_struc}{name_file}_transferred_to_swiss_model.tsv.gz", sep = '\t' )
-            dict_sp_swiss = __Parse_Structural_Properties__( df_sp_swiss, 1 ) # parse structural data from swiss-model
-
-            dict_sp = __Combine_Structural_Properties__( dict_sp_swiss, dict_sp ) # combine structural property data from RCSB_PDB and SWISS-MODEL
-
-        ''' initialize arrays of proteins with no aligned structures '''
-        for id_protein in dict_fasta_protein :
-            len_seq = len( dict_fasta_protein[ id_protein ] ) # retrieve length of protein
-            for str_datatype in [ 'acc', 'phi', 'psi', 'ss8', 'structure_id', 'datatype_acc' ] :
-                if id_protein not in dict_sp[ str_datatype ] :
-                    ''' set default array for each datatype '''
-                    if str_datatype == 'datatype_acc' :
-                        arr = np.full( len_seq, 0, dtype = float )
-                    elif str_datatype == 'structure_id' :
-                        arr = np.full( len_seq, np.nan, dtype = object )
-                    else :
-                        arr = np.full( len_seq, np.nan )
-                    dict_sp[ str_datatype ][ id_protein ] = arr
-
-        """
-        Prediction of Structural Properties Using Machine-Learning Models
-        """
-        # 2021-05-29 23:41:46 
-        # use machine-learning model to predict structural properties
-        ''' settings '''
-        int_length_flanking = 74 # number of protein residues flanking the residue whose structural property values are predicted
-
-        ''' download models '''
-        file_name_model_rsa = "RSA.Conv1D.5_large_and_wide_dense.epoch_082.val_loss_0.1185.hdf5" # name of machine learning model used for structural property prediction (RSA)
-        file_name_model_ss8 = "SS8.Conv1D.5_large_and_wide_dense.epoch_031.val_loss_0.7378.hdf5" # name of machine learning model used for structural property prediction (SS8)
-
-        PKG.Download_Data( f"structural_property_estimation/{file_name_model_rsa}", dir_remote, name_package ) # download data
-        PKG.Download_Data( f"structural_property_estimation/{file_name_model_ss8}", dir_remote, name_package ) # download data
-
-        ''' prepare encoding '''
-        def Index_List( l, start = 0 ) :
-            ''' # 2021-05-06 22:32:53 
-            index a given list containing hashable element with integer index starting with 'start' (default: 0)
-            can be used for 'one-hot encoding'
-            '''
-            return dict( ( e, i ) for e, i in zip( l, np.arange( start, len( l ) + start, 1, dtype = int ) ) )
-
-        # input structure : amino-acids (one-hot encoding) + ss8 (one-hot encoding) + rsa (0~1)
-        int_index_start_amino_acid = 0
-        l_amino_acids = [ 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y' ]
-        dict_amino_acid_to_index = Index_List( l_amino_acids, int_index_start_amino_acid ) # one-hot encoding dictionary for 20 amino acids
-        int_index_start_ss8 = len( l_amino_acids )
-        l_ss8 = [ 'G', 'H', 'I', 'E', 'B', 'T', 'S', 'C' ]
-        dict_ss8_to_index = Index_List( l_ss8, int_index_start_ss8 ) # one-hot encoding dictionary for 20 amino acids
-        index_rsa = len( l_amino_acids ) + len( l_ss8 ) # index of rsa data in the input array
-
-
-        ''' predict and update structural properties using ML models for each batch '''
-        import tensorflow as tf
-        from tensorflow.keras import layers
-
-        l_id_protein_all = list( dict_sp[ 'acc' ] ) # retrieve a list containing all 'id_protein'
-        n_batches = int( np.ceil( len( l_id_protein_all ) / int_number_of_proteins_in_a_batch_during_dnn_prediction ) ) # number of batches
-        ''' for each batch '''
-        for l_id_protein in LIST_Split( l_id_protein_all, n_split = n_batches ) :
-            ''' perform encoding for the current batch '''
-            # 'l_id_protein': id_protein -> integer encoding to reduce memory usage
-            # perform encoding and record metadata of each position
-            x_data = [ ]
-            meta_data = [ ]
-
-            for index_protein, id_protein in enumerate( l_id_protein ) :
-                seq = dict_fasta_protein[ id_protein ]
-                length = len( seq ) # retrieve the length of the protein
-
-                arr_acc = dict_sp[ 'acc' ][ id_protein ]
-                arr_ss8 = dict_sp[ 'ss8' ][ id_protein ]
-
-                ''' one-hot encoding (+ RSA) of the current protein '''
-                arr_protein = np.zeros( ( length, len( l_amino_acids ) + len( l_ss8 ) + 1 ), dtype = np.float32 ) # initialize one-hot encoding of the proteinarray
-                for i in range( length ) :
-                    # one-hot encoding of amino acid
-                    aa = seq[ i ] # retrieve amino acid
-                    if aa in dict_amino_acid_to_index : # invalid amino acid will be considered as a missing value (empty row)
-                        arr_protein[ i, dict_amino_acid_to_index[ aa ] ] = 1 
-                    # one-hot encoding of ss8
-                    if not np.isnan( arr_ss8[ i ] ) : # missing value = empty row
-                        arr_protein[ i, dict_ss8_to_index[ l_ss8[ int( arr_ss8[ i ] ) ] ] ] = 1
-                    # encode rsa information
-                    if not np.isnan( arr_acc[ i ] ) : # missing value = empty row
-                        arr_protein[ i, index_rsa ] = arr_acc[ i ]
-
-                ''' prepare one-hot encoding input array for each missing position '''
-                ''' assumes positions of missing values in 'arr_acc' are identical to those of 'arr_ss8' '''
-                for int_pos in np.where( np.isnan( arr_acc ) )[ 0 ] : 
-                    start = max( 0, int_pos - int_length_flanking )
-                    end = min( length, int_pos + int_length_flanking + 1 )
-
-                    arr = np.zeros( ( int_length_flanking + 1 + int_length_flanking, len( l_amino_acids ) + len( l_ss8 ) + 1 ), dtype = np.float32 ) # initialize input array
-                    start_input = max( 0, int_length_flanking - int_pos ) # index of start position of the input array
-                    arr[ start_input : start_input + end - start ] = arr_protein[ start : end ]
-                    x_data.append( arr )
-                    meta_data.append( [ index_protein, int_pos ] )
-            x_data = np.array( x_data, dtype = np.float32 )
-
-            ''' predict and update structural properties using ML models for the current batch '''
-            for dir_file_model, str_datatype in zip( [ f"{dir_folder_cressp}structural_property_estimation/{file_name_model_rsa}", f"{dir_folder_cressp}structural_property_estimation/{file_name_model_ss8}" ], [ 'acc', 'ss8' ] ) :
-
-                model = tf.keras.models.load_model( dir_file_model )
-
-                with tf.distribute.MirroredStrategy( ).scope( ) : # distributed calculation
-                    y_pred = model.predict( x_data ) # predict structural properties using a ML model
-                    y_pred = y_pred.argmax( axis = 1 ) if str_datatype == 'ss8' else y_pred # multiclass classification for 'ss8' and regression for 'acc'
-
-                for i in range( len( y_pred ) ) :
-                    index_protein, int_pos = meta_data[ i ] # parse metadata
-                    id_protein = l_id_protein[ index_protein ]
-                    dict_sp[ str_datatype ][ id_protein ][ int_pos ] = y_pred[ i ] # update structural property data with predicted data
-
-        """
-        Save Result
-        """
-        def __Encode_Structural_Properties__( dict_sp, dict_fasta_protein ) :
-            """
-            # 2021-05-29 23:56:59 
-            Compose a dataframe containing encoded structural properties using given 'dict_sp' and 'dict_fasta_protein'
-
-            'dict_sp' : dictionary containing structural properties
-            'dict_fasta_protein' : dictionary containing protein sequences
-
-            """
-            dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
-            dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
-            dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
-            dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )
-
-            ''' initialize dataframe with protein sequences '''
-            df_sp = pd.Series( dict_fasta_protein, name = 'seq' ).reset_index( ).rename( columns = { 'index' : 'id_protein' } ).set_index( 'id_protein' )
-
-            # encode combined structural properties into ASCII strings using ASCII encoding
-            df_sp[ 'rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1' ] = pd.Series( ASCII_Encode( dict_sp[ 'acc' ], ** dict_kw_rsa ) )
-            df_sp[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ] = pd.Series( ASCII_Encode( dict_sp[ 'phi' ], ** dict_kw_torsion_angle ) )
-            df_sp[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ] = pd.Series( ASCII_Encode( dict_sp[ 'psi' ], ** dict_kw_torsion_angle ) )
-            df_sp[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ] = pd.Series( ASCII_Encode( dict_sp[ 'ss8' ], ** dict_kw_ss8 ) )
-            df_sp[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ] = pd.Series( ASCII_Encode( dict_sp[ 'datatype_acc' ], ** dict_kw_datatype ) )
-            df_sp[ 'structure_id___redundancy_reduced' ] = pd.Series( dict( ( h, Encode_List_of_Strings( dict_sp[ 'structure_id' ][ h ] ) ) for h in dict_sp[ 'structure_id' ] ) )
-            df_sp.reset_index( drop = False, inplace = True )
-            return df_sp
-
-        df_sp = __Encode_Structural_Properties__( dict_sp, dict_fasta_protein ) # encode structural properties into a dataframe
-        df_sp[ 'fasta_header' ] = list( dict_header[ i ] for i in df_sp.id_protein.values ) # retrieve headers
-        df_sp.to_csv( f"{dir_folder_pipeline}{name_file}.tsv.gz", sep = '\t', index = False ) # save structural properties of input proteins as a tabular file
+        __Combine_Structural_Properties_and_Predict_Remaining_Residues__( dir_file_protein, dir_folder_pipeline, dir_folder_pipeline_temp, flag_use_rcsb_pdb_only, int_number_of_proteins_in_a_batch_during_dnn_prediction, flag_use_all_gpu_devices ) # predict RSA
+            
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
